/*
 * Program to test simulation of page replacement.
 *
 * Input consists of command-line arguments and an input file:
 *
 * Command-line arguments:
 * Required:
 *   name of input file (format below)
 *   number of page frames
 * Optional:
 *   "--debug" to produce extra debugging information
 *   "--interval N" to specify interval for "clock ticks", for algorithms 
 *      that need this (the idea being to consider that a "clock tick" happens 
 *      every N references)
 *   "--tau N" to specify time interval for working set algorithms
 *
 * Input file format:
 * number of pages
 * one or more lines of the form "t R n" or "t W n", where"
 *   t is the time of access
 *   R/W indicates whether this is a read or write reference
 *   n is the page number being referenced
 *
 * Output:  for each page replacement algorithm implemented, the following 
 *     information:
 * total number of page references
 * number of page references that changed the page ('W')
 * number of page faults
 * number of times a page had to be written out
 *
 * We assume that all page references are valid -- if the page is not in
 * memory, it's on disk.
 *
 * If the optional "--debug" command-line argument is given, simulations 
 * can also print additional information that might be helpful in debugging.
 */
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <vector>

using std::ifstream;
using std::vector;

#include "page-reference.hpp"
#include "results.hpp"

bool DEBUG = false;

#include "pra-Base.hpp"

#include "pra-Aging.cpp"
#include "pra-Clock.cpp"
#include "pra-FIFO.cpp"
#include "pra-LRU.cpp"
#include "pra-NFU.cpp"
#include "pra-NRU.cpp"
#include "pra-Optimal.cpp"
#include "pra-SecondChance.cpp"
#include "pra-WSClock.cpp"
#include "pra-WorkingSet.cpp"

//
// function declarations
//

// exits program on error
long getPositive(const char *inString, const char *msg);

// exits program on error
void parseInputFile(const char *filename, long &pages,
        vector<PageReference> &refs);

// main program

int main(int argc, char *argv[]) {
    const char *usageMsg = "usage %s:  "
        "infile frames " 
        "[--interval n] "
        "[--tau n] "
        "[--debug] "
        "\n";

    // process arguments 

    // required arguments

    if (argc < 3) {
        fprintf(stderr, usageMsg, argv[0]);
        exit(EXIT_FAILURE);
    }

    char *inFileName = argv[1];

    long numFrames = getPositive(argv[2], 
            "numFrames must be positive integer");

    // optional arguments

    long interval = 10;
    long tau = 10;

    int argi = 3;
    while (argi < argc) {
        if (strcmp(argv[argi], "--debug") == 0) {
            DEBUG = true;
            argi+=1;
        } else if (strcmp(argv[argi], "--interval") == 0) {
            argi+=1;
            if (argi >= argc) {
                fprintf(stderr, usageMsg, argv[0]);
                exit(EXIT_FAILURE);
            }
            interval = getPositive(argv[argi],
                    "clock tick must be positive integer");
            argi+=1;
        } else if (strcmp(argv[argi], "--tau") == 0) {
            argi+=1;
            if (argi >= argc) {
                fprintf(stderr, usageMsg, argv[0]);
                exit(EXIT_FAILURE);
            }
            tau = getPositive(argv[argi],
                    "tau must be positive integer");
            argi+=1;
        } else {
            fprintf(stderr, usageMsg, argv[0]);
            exit(EXIT_FAILURE);
        }
    }

    // read input

    vector<PageReference> refs;
    long numPages;
    parseInputFile(inFileName, numPages, refs);

    // print some starter information

    printf("Number of frames = %ld\n", numFrames);
    printf("Number of pages = %ld\n", numPages);
    printf("Interval to clear R bit = %ld\n", interval);
    printf("Tau = %ld\n", tau);
    puts("");

    if (DEBUG) {
        printf("References:\n");
        printf("%ld\n", numPages);
        for (auto ref : refs) {
            printf("%s\n", ref.toString().c_str());
        }
        puts("");
    }

    // perform simulations and print results

    vector <PRA *> praAlgorithms;

    // make list of all algorithms to be simulated
    praAlgorithms.push_back(new PRA_FIFO());
    praAlgorithms.push_back(new PRA_Optimal());
    praAlgorithms.push_back(new PRA_SecondChance());
    praAlgorithms.push_back(new PRA_Clock());
    praAlgorithms.push_back(new PRA_NRU());
    praAlgorithms.push_back(new PRA_LRU());
    praAlgorithms.push_back(new PRA_NFU());
    praAlgorithms.push_back(new PRA_Aging());
    praAlgorithms.push_back(new PRA_WorkingSet());
    praAlgorithms.push_back(new PRA_WSClock());

    // simulate them and print results
    for (auto pra : praAlgorithms) {
        Results r = pra->simulate( numFrames, numPages, interval, tau, refs);
        if (r.isValid) {
            printf("Results for %s:\n", pra->getName().c_str());
            puts("");
            r.print();
            puts("");
        }
    }

    // cleanup
    for (auto &pra : praAlgorithms) {
        delete pra;
    }

    return EXIT_SUCCESS;
}

// 
// function definitions
// 

long getPositive(const char *inString, const char *msg) {
    char *endPtr;
    long n = strtol(inString, &endPtr, 10);
    if (*endPtr != '\0') {
        fprintf(stderr, "%s\n", msg);
        exit(EXIT_FAILURE);
    }
    return n;
}

void parseInputFile(const char *filename, long &pages,
        vector<PageReference> &refs)
{
    ifstream infile;
    infile.open(filename);
    if (!infile.is_open()) {
        fprintf(stderr, "cannot open input file %s\n", filename);
        exit(EXIT_FAILURE);
    }
    if (!(infile >> pages) || (pages <= 0)) {
        fprintf(stderr, "input error reading num pages\n");
        exit(EXIT_FAILURE);
    }

    long time;
    char ch;
    long num;

    long lineNumber = 1;
    long prevTime = -1;
    while (infile >> time >> ch >> num) {
        lineNumber += 1;
        if (time <= prevTime) {
            fprintf(stderr, "input error on line %ld\n", lineNumber);
            exit(EXIT_FAILURE);
        }
        prevTime = time;
        if (num >= pages) {
            fprintf(stderr, "input error on line %ld\n", lineNumber);
            exit(EXIT_FAILURE);
        }
        if (ch == 'R') {
            refs.push_back(PageReference(time, num, false));
        } else if (ch == 'W') {
            refs.push_back(PageReference(time, num, true));
        } else {
            fprintf(stderr, "input error on line %ld\n", lineNumber);
            exit(EXIT_FAILURE);
        }
    }
    if (!infile.eof()) {
        fprintf(stderr, "input error on line %ld\n", lineNumber);
        exit(EXIT_FAILURE);
    }
    infile.close();
}

