#include <cstdlib>
#include <cassert>

#include "pra-Base.hpp"

#include "page-reference.hpp"
#include "results.hpp"

#include "pte.hpp"

#include <string>
#include <vector>
// HINT:  you probably will not need this.
// really we only need a regular queue, but this "double-ended" queue 
//   permits printing its elements.
#include <deque>

using std::string;
using std::vector;
using std::deque;

//
// notes:
// 
// this class is meant to be not only an implementation of the FIFO
// algorithm but also something of a model for other classes.  so it
// sometimes does more than it needs to (e.g., defining what should
// happen on clock ticks -- nothing here but some other algorithms need
// to do something).
//
class PRA_FIFO : public PRA {

    private:

        // variables to be used in simulation 
        // (make them member variables so all methods have access)

        // FIXME?  okay to remove this but I think it's useful
        Results results;

        // track whether each frame is in use 
        // HINT:  
        // you probably want something like this.
        vector<bool> framesInUse;

        // page table
        // HINT:  
        // you almost sure want something like this, though for some
        // algorithms you may want to make a subclass of PTE that includes 
        // additional fields.  more in the comments for PTE.
        vector<PTE> pageTable;

        // queue of pages, in order of loading
        // HINT:
        // you probably want to replace this with something appropriate
        // to the algorithm you're implementing -- an array of counters
        // for Aging, for example.
        deque<long> pageQueue;
        
    public:

        // methods required by superclass

        // provide name of algorithm
        string getName(void) const;

        // simulate operation of algorithm 
        Results simulate(
                const long numFrames,
                const long numPages,
                const long interval, 
                const long tau,
                const vector<PageReference> &refs);
    private:

        // methods for this class
        // HINT:
        // I think it's useful to structure the simulation in terms of
        // these methods, but you can do something else if you like.

        // simulate operation of MMU:
        // update R and M bits 
        // return frame number if entry is valid, else -1 ("page fault")
        long simulateMMU(const long referencedPage, const bool isWrite);

        // define what happens on periodic clock ticks (e.g., clear R bits)
        void onClockTick(void);

        // define what happens on page fault 
        void onPageFault(const long referencedPage);

        // return index of first free frame, or -1 if none found
        long findFreeFrame(void);

        // return index of page to steal
        long findPageToSteal(void);

        // print page queue (for debugging)
        // HINT:
        // you probably do *not* need this method.
        void printPageQueue(void);
};

// methods required by superclass

string PRA_FIFO::getName(void) const { return "FIFO"; }

Results PRA_FIFO::simulate(
        const long numFrames,
        const long numPages,
        const long interval,
        const long tau,
        const vector<PageReference> &refs)
{
    if (DEBUG) {
        printf("simulating %s\n", getName().c_str());
    }

    // HINT:
    // here is where you can initialize member variables.

    framesInUse = vector<bool>(numFrames, false);
    pageTable = vector<PTE>(numPages);

    results.clear();

    // HINT:
    // you can probably reuse most or even all of the rest of the code
    // in this method, except for anything that uses pageQueue.

    // index into "refs" input
    long nextReference = 0;

    // loop over times -- this allows handling both periodic clock
    // ticks and page references
    long lastTime = refs[refs.size()-1].time;
    for (long t = 0; t <= lastTime; ++t) {

        // periodic clock tick

        if (((t % interval) == 0) && (t > 0)) {
            if (DEBUG) {
                printf("at time %ld clock tick\n", t);
            }
            onClockTick();
        }

        // page reference

        if (t == refs[nextReference].time) {
            PageReference ref = refs[nextReference++];
            results.refs += 1;
            if (ref.isWrite) {
                results.writeRefs += 1;
            }

            // simulate operation of MMU
            long frame = simulateMMU(ref.pageNumber, ref.isWrite);
            if (DEBUG) {
                if (frame >= 0) {
                    printf("at time %ld page %ld is in frame %ld\n",
                            t, ref.pageNumber, frame);
                } else {
                    printf("at time %ld page fault for page %ld\n",
                            t, ref.pageNumber);
                }
            }

            // page fault?
            if (frame < 0) {
                onPageFault(ref.pageNumber);
                frame = simulateMMU(ref.pageNumber, ref.isWrite);
                assert(frame >= 0);
            }
        }
    }

    if (DEBUG) { puts(""); }
    results.isValid = true;
    return results;
}

// methods for this class

// simulate operation of MMU:
// update R and M bits 
// return frame number if entry is valid, else -1 ("page fault")
// HINT:
// you may be able to copy this method unchanged, except for the class name.
long PRA_FIFO::simulateMMU(const long referencedPage, const bool isWrite) {
    if (pageTable[referencedPage].valid) {
        pageTable[referencedPage].rBit = true;
        if (isWrite) {
            pageTable[referencedPage].mBit = true;
        }
        return pageTable[referencedPage].frame;
    } else {
        return -1;
    }
}


// define what happens on periodic clock ticks (e.g., clear R bits)
// HINT:
// for some algorithms (Aging for example) you will need to add to this,
// but most can use it as is.
void PRA_FIFO::onClockTick(void) {
    // clear R bits
    for (auto &p : pageTable) {
        p.rBit = false;
    }
}

// define what happens on page fault 
// HINT:
// you can probably reuse almost all of this method, all but references
// to page queue.
void PRA_FIFO::onPageFault(const long referencedPage) {
    results.pageFaults += 1;
    long frame = findFreeFrame();
    if (frame == -1) {
        // no free frames -- find one to steal
        if (DEBUG) {
            printf("no free frames\n");
            PTE::printPageTable(pageTable);
            printPageQueue();
        }
        long toSteal = findPageToSteal();
        assert((toSteal >= 0) && (toSteal < (long) pageTable.size()));
        assert(pageTable[toSteal].valid);
        frame = pageTable[toSteal].frame;
        if (DEBUG) {
            printf("removing page %ld from frame %ld\n", toSteal, frame);
        }
        // got one -- need to write it out?
        if (pageTable[toSteal].mBit) {
            if (DEBUG) {
                printf("writing out page %ld\n", toSteal);
            }
            results.pageWriteOuts += 1;
        }
        pageTable[toSteal].valid = false;
    } 
    // simulate loading page
    if (DEBUG) {
        printf("loading page %ld into frame %ld\n", referencedPage, frame);
    }
    framesInUse[frame] = true;
    pageTable[referencedPage].valid = true;
    pageTable[referencedPage].frame = frame;
    pageTable[referencedPage].rBit = true;
    pageTable[referencedPage].mBit = false;
    // HINT:
    // you probably need to change this.
    pageQueue.push_back(referencedPage);
}

// methods for this class

// return index of first free frame, or -1 if none found
// HINT:
// you may be able to copy this method unchanged, except for the class name.
long PRA_FIFO::findFreeFrame(void) {
    for (size_t i = 0; i < framesInUse.size(); ++i) {
        if (!framesInUse[i]) {
            return i;
        }
    }
    return -1;
}

// return index of page to steal
// for this algorithm, choose the one at the start of the queue.
long PRA_FIFO::findPageToSteal(void) {
    assert(!pageQueue.empty());
    long pageToSteal = pageQueue.front();
    pageQueue.pop_front();
    return pageToSteal;
}

// print page queue
// HINT:
// you probably will not need this method.
void PRA_FIFO::printPageQueue(void) {
    printf("page queue:  ");
    for (auto p : pageQueue) {
        printf("%ld ", p);
    }
    printf("\n");
}
